/* ccgo: sgf/go_node.cc
 * 
 * Copyright 2002 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "go_node.hh"
#include "../setup.hh"
#include "../move.hh"
#include "../put_move.hh"
#include "../handicap.hh"
#include "../proper_node.hh"

using namespace sgf;
using namespace std;

go::Node * GoNode::sub_game_tree(go::Game & g) const
{
	go::ProperNode * pn = new go::ProperNode;
	go::ModifiProper * mp = new go::ModifiProper;
	setup(* mp, g);
	pn->set_proper(mp);

	const go::Modifi * m = mp->get_modifi();

	if (m) m->apply(g);
	for (unsigned i = 0; i < get_link_size(); i ++) {
		pn->set_next(dynamic_cast<GoNode *>(get_next(i))->sub_game_tree(g));
	}
	if (m) m->undo(g);

	return pn;
}

GoNode::GoNode(const sgf::Node & n) :
	sgf::Node(n)
{
}

GoNode::GoNode(const Tree & t) :
	sgf::Node(t.get_node_list()[0])
{
	GoNode * p = this;

	for (vector<sgf::Node>::const_iterator i = t.get_node_list().begin() + 1; i  != t.get_node_list().end(); i ++) {
		GoNode * pp = new GoNode(* i);
		p->set_next(pp);
		p = pp;
	}

	for (vector<Tree>::const_iterator i = t.get_sub_trees().begin(); i  != t.get_sub_trees().end(); i ++) {
		p->set_next(new GoNode(* i));
	}
}

GoNode::~GoNode()
{
}

void GoNode::setup(go::Proper & p, unsigned z) const
{
	const sgf::Prop * pp;

	if (pp = find("N")) p.set_name(pp->as_string());
	else p.set_name("");

	if (pp = find("C")) p.set_comment(pp->as_string());
	else p.set_comment("");

	if (pp = find("LB")) p.set_labels(pp->ar_label(z));
	else p.set_labels(std::vector<go::Proper::Label>());

	vector<go::Proper::Marking> marks;
	if (pp = find("CR")) {
		std::vector<go::Loc> l = pp->ar_loc(z);
		for (std::vector<go::Loc>::const_iterator i = l.begin(); i != l.end(); i ++) marks.push_back(go::Proper::Marking(* i, go::MARK_CIRCLE));
	}
	if (pp = find("TR")) {
		std::vector<go::Loc> l = pp->ar_loc(z);
		for (std::vector<go::Loc>::const_iterator i = l.begin(); i != l.end(); i ++) marks.push_back(go::Proper::Marking(* i, go::MARK_TRIANGLE));
	}
	if (pp = find("SQ")) {
		std::vector<go::Loc> l = pp->ar_loc(z);
		for (std::vector<go::Loc>::const_iterator i = l.begin(); i != l.end(); i ++) marks.push_back(go::Proper::Marking(* i, go::MARK_SQUARE));
	}
	if (pp = find("MA")) {
		std::vector<go::Loc> l = pp->ar_loc(z);
		for (std::vector<go::Loc>::const_iterator i = l.begin(); i != l.end(); i ++) marks.push_back(go::Proper::Marking(* i, go::MARK_CROSS));
	}
	if (pp = find("SL")) {
		std::vector<go::Loc> l = pp->ar_loc(z);
		for (std::vector<go::Loc>::const_iterator i = l.begin(); i != l.end(); i ++) marks.push_back(go::Proper::Marking(* i, go::MARK_SELECT));
	}
	if (pp = find("TW")) {
		std::vector<go::Loc> l = pp->ar_loc(z);
		for (std::vector<go::Loc>::const_iterator i = l.begin(); i != l.end(); i ++) marks.push_back(go::Proper::Marking(* i, go::MARK_W_TERRITORY));
	}
	if (pp = find("TB")) {
		std::vector<go::Loc> l = pp->ar_loc(z);
		for (std::vector<go::Loc>::const_iterator i = l.begin(); i != l.end(); i ++) marks.push_back(go::Proper::Marking(* i, go::MARK_B_TERRITORY));
	}
	p.set_marks(marks);
}

void GoNode::setup(go::RootProper & p) const
{
	const sgf::Prop * pp;
	if (pp = find("SZ")) p.set_size(pp->as_int());
	if (pp = find("KM")) p.set_komi(pp->as_real());
	if (pp = find("GN")) p.set_game_name(pp->as_string());
	setup((go::Proper &) p, p.get_size());
}

void GoNode::setup(go::ModifiProper & p, const go::Game & g) const
{
	setup((go::Proper &) p, g.get_board_size());
	const sgf::Prop * pp;
	std::vector<go::Setup::Site> s;
	bool any_setup = false;
	// checking if there is any "setup" operations
	if (pp = find("AB")) {
		std::vector<go::Loc> l = pp->ar_loc(g.get_board_size());
		for (std::vector<go::Loc>::iterator i = l.begin(); i != l.end(); i ++) s.push_back(go::Setup::Site(* i, go::STATE_BLACK));
		if (! find("HA")) any_setup = true;
	}
	if (pp = find("AW")) {
		std::vector<go::Loc> l = pp->ar_loc(g.get_board_size());
		for (std::vector<go::Loc>::iterator i = l.begin(); i != l.end(); i ++) s.push_back(go::Setup::Site(* i, go::STATE_WHITE));
		any_setup = true;
	}
	if (pp = find("AE")) {
		std::vector<go::Loc> l = pp->ar_loc(g.get_board_size());
		for (std::vector<go::Loc>::iterator i = l.begin(); i != l.end(); i ++) s.push_back(go::Setup::Site(* i, go::STATE_EMPTY));
		any_setup = true;
	}
	go::Turn tn;
	if (pp = find("PL")) {
		if (pp->as_char() == 'B') tn = go::TURN_BLACK;
		else if (pp->as_char() == 'W') tn = go::TURN_WHITE;
		else { // impossible???
			tn = go::TURN_BLACK;
			std::cerr << "Turn error!!!" << std::endl;
		}
		any_setup = true;
	} else tn = g.get_turn();
	if (any_setup) {
		// in FF[4] B and W are not allowed in a setup node
		if (pp = find("B")) {
			go::Loc l = pp->as_loc(g.get_board_size());
			if (l != go::Loc()) s.push_back(go::Setup::Site(l, go::STATE_BLACK));
		}
		if (pp = find("W")) {
			go::Loc l = pp->as_loc(g.get_board_size());
			if (l != go::Loc()) s.push_back(go::Setup::Site(l, go::STATE_WHITE));
		}
		go::Setup * setup = new go::Setup(s, tn);
		setup->check(g);
		p.set_modifi(setup);
		return;
	}

	// try building a "move" modifi
	go::Move * move = 0;
	if (pp = find("B")) {  // only one of B, W, and HA should happen
		go::Loc l = pp->as_loc(g.get_board_size());
		if (l == go::Loc()) move = new go::Move(go::TURN_BLACK); // B pass
		else move = new go::PutMove(go::TURN_BLACK, l); // put B
		// std::cerr << "Black move node\n";
	} else if (pp = find("W")) {
		go::Loc l = pp->as_loc(g.get_board_size());
		if (l == go::Loc()) move = new go::Move(go::TURN_WHITE); // W pass
		else move = new go::PutMove(go::TURN_WHITE, l); // put W
		// std::cerr << "White move node\n";
	} else if (pp = find("HA")) { // handicap
		int h = pp->as_int();
		if (pp = find("AB")) move = new go::Handicap(pp->ar_loc(g.get_board_size())); // w/ handicap positions
		else move = new go::Handicap(h); // no handicap positions
	}

	if (move) {
		if (pp = find("MN")) move->set_num(pp->as_int());
		if (move->check(g)) p.set_modifi(move);
		else delete move;
	}
}

go::Node * GoNode::get_game_tree() const
{
	go::ProperNode * pn = new go::ProperNode;
	go::RootProper * rp = new go::RootProper;
	setup(* rp);
	pn->set_proper(rp);

	go::Game g;
	g.set_board_size(rp->get_size());
	g.set_komi(rp->get_komi());
	g.set_name(rp->get_game_name());
	g.set_rule(go::rule_sgf);
	for (unsigned i = 0; i < get_link_size(); i ++) {
		pn->set_next(dynamic_cast<GoNode *>(get_next(i))->sub_game_tree(g));
	}
	return pn;
}
